/*   Gram -- A graph motif solver.
 *
 *   (C) 2008-2009 René van Bevern <m5bere2@uni-jena.de>
 *
 *   This file is part of Gram.
 *
 *   Gram is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Gram is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Gram.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "GraphMotifInstance.hpp"
#include "GraphUtils.hpp"
#include "SubsetIterator.hpp"
#include <boost/unordered_set.hpp>
#include <boost/variant.hpp>
#include <boost/math/special_functions.hpp>
#include <numeric>

using namespace std;

inline ColorfulGraphMotifInstance::
ColorfulGraphMotifInstance(Graph& _G, ColorSet _M, int _k, ColorSet _L)
  : G(_G), M(_M), L(_L), k(_k), coloring(get(vertex_color, G)), lists(get(vertex_rank, G)) {
  VertexIterator v0, vn, v;
  remove_edges(G, unicolored_edge);
  if(k > 1) remove_vertices(G, isolated);
  tie(v0, vn) = vertices(G);
    RankedSubsetIterator<ColorSet> end, c(M, 2);
    for(; c != end; ++c)
      for(v = v0; v != vn; ++v)
	if(member(coloring[*v], *c) && !(*c & ColorSet(lists[*v])).empty())
	  table.write(*v, *c, Leaf());
}

inline int ColorfulGraphMotifInstance::growMotifs(Vertex v, const ColorSet& C)
{
  ColorSet C_(C), clist(get(vertex_rank,G,v)); C_.erase(coloring[v]);
  Graph::out_edge_iterator e0, en;
  VertexWeight w_opt = -1;
  Vertex v_opt;
  ColorSet vc_opt;
  Edge e_opt;
  RankedSubsetIterator<ColorSet> end;
  
  for(tie(e0, en) = out_edges(v, G); e0 != en; e0++) {
    Vertex a0 = target(*e0, G);
    for(RankedSubsetIterator<ColorSet> C2(clist,1); C2!=end; ++C2) {
      if(table(a0, C_-*C2)) {
	VertexWeight w = table.weight(a0, C_-*C2) + get(edge_weight, G, *e0);
	if(w > w_opt) {
	  v_opt = a0;
	  vc_opt = *C2;
	  w_opt = w;
	  e_opt = *e0;
	}
      }
    }
  }
  if (w_opt >= 0) {
    table.write(v, C, Add(v_opt, vc_opt), w_opt);
    return w_opt;
  } else return -1;
}
  
inline bool ColorfulGraphMotifInstance::joinMotifs(Vertex v, const ColorSet& C, int thres)
{
    SubsetIterator<ColorSet> end, i(C);
    Color colV(coloring[v]);
    ColorSet C_, C__, clist(get(vertex_rank,G,v));
    VertexWeight w_opt = -1;
    ColorSet C_opt, C__opt;
    
    for(++i; i != end; ++i) {
      if((*i-L).size() == (*i&L).size())
	for(RankedSubsetIterator<ColorSet> C2(clist,1); C2!=end; ++C2) {
	  C_ = *i | *C2; C__ = (C - C_) | *C2;
	  C_.insert(colV); C__.insert(colV);
	  if(table(v, C_) and table(v, C__)) {
	    VertexWeight w1 = table.weight(v, C_);
	    VertexWeight w2 = table.weight(v, C__);
	    if(w1 + w2 > w_opt) {
	      C_opt = C_;
	      C__opt = C__;
	      w_opt = w1 + w2;
	    }
	  }
	}
    }
    
      if(w_opt > thres)
	table.write(v, C, Join(C_opt, C__opt), w_opt);
      if (thres >= 0 || w_opt >= 0) return true;
    return false;
}

inline tuple<VertexWeight, Solution> ColorfulGraphMotifInstance::solve() {
  RankedSubsetIterator<ColorSet> end;
  int relevantCount = 0, k_;
  VertexIterator v0, v, vn;
  
  for(k_ = 4; k_ <= k; k_+=2,relevantCount = 0) {
    RankedSubsetIterator<ColorSet> C(M, k_);
    unordered_set<Vertex> toKeep;
    for(tie(v0, vn) = vertices(G); C != end; ++C)
	for(v = v0; v != vn; ++v) {
	  if(member(coloring[*v], *C) && !(*C & ColorSet(get(vertex_rank, G, *v))).empty()
	     && ((*C-L).size() == (*C&L).size())) {
	      int w = growMotifs(*v, *C);
	      if(joinMotifs(*v, *C, w))
		if(toKeep.insert(*v).second) ++relevantCount;
	  }
	}
    
    if(2*relevantCount < k) break;
    remove_vertices(G, is_not_in<unordered_set<Vertex> >(toKeep));	
  }

  VertexWeight w_opt = -1;
  Vertex v_opt = 0;
  ColorSet C_opt;

  for(tie(v, vn) = vertices(G); v != vn; ++v) {
    for(RankedSubsetIterator<ColorSet> C(M, k_-2); C != end; ++C) {
      if((*C-L).size() == (*C&L).size() && table(*v, *C)) {
	VertexWeight w = table.weight(*v, *C);
	if(w > w_opt) {
	  w_opt = w;
	  v_opt = *v;
	  C_opt = *C;
	}
      }
    }
  }
  if(w_opt >= 0) return make_tuple(w_opt, backtrack(v_opt, G, C_opt, table));
  else return make_tuple(-1, Solution());
}

int numcolors(set<Vertex> L, Graph &G) {
  ColorSet allcolors;

  for(set<Vertex>::iterator i = L.begin(); i != L.end(); ++i)
    allcolors = allcolors | ColorSet(get(vertex_rank, G, *i));

  return (allcolors.size() < L.size()) ? allcolors.size() : L.size();
}

int seed() {
  FILE *f = fopen("/dev/urandom", "r");
  int res;
  if(!f) res = time(0);
  else {
    res = getc(f);
    res = (res << 8) | getc(f);
  }
  if(f) fclose(f);
  cout << "random seed               : " << res << endl;
  return res;
}

GraphMotifInstance::GraphMotifInstance(const Graph& _G, const MultiColorSet& _M, float err)
  : G(_G), k(_M.size()), coloring(get(vertex_color, G)), errorProb(err), RNG(seed()) {
  VertexIterator v0, vn;
  
  M = ColorSet(_M.begin(),_M.end());
  L = genLabels(_M);
  if(k > 1) remove_vertices(G, isolated);
  Ls = colorset_intersection_graph(G);
  colorsPerL = vector<int>(Ls.size());
  cout << "labels for indep. compon. : ";
  for(int i = 0; i < Ls.size(); ++i) {
    colorsPerL[i] = numcolors(Ls[i], G);
    cout << colorsPerL[i] << " ";
  }
  cout << endl;
}

inline void GraphMotifInstance::recolor() {
  int curlabel = 0;
  for(int i = 0; i < Ls.size(); ++i) {
    for(set<Vertex>::iterator j = Ls[i].begin(); j != Ls[i].end(); ++j)
      coloring[*j] = L[curlabel + getRand(colorsPerL[i])];
    curlabel += colorsPerL[i];
  }
}

unsigned long int trials(const vector<int> &colorsPerL, float e) {
  double hit = 1;
  for(int i = 0; i < colorsPerL.size(); ++i) {
    hit *= math::factorial<double>(colorsPerL[i])
      / pow((double)colorsPerL[i],colorsPerL[i]);
  }
  return (long unsigned int)ceil(log(e)/log(1-hit));
}

Solution GraphMotifInstance::solve() {
  int t = trials(colorsPerL, errorProb);
  if(M.empty()) return Solution();

  ColorSets rep;
  VertexWeight w_opt = -1;
  int w_size = 0;
  
  cout << "trials                    : " << t << endl;
  cout << "searching for             : ";
  M.printSet();
  cout << endl;
  
  ColorSet LL(L.begin(), L.end());
  M = M | LL;
  k = M.size();
  cout << "search motif size         : " << M.size() << endl;

  Solution r_opt;
  for(int i = 0; i <= t; ++i) {
    recolor();
    Solution result;
    VertexWeight w = -1;
      
    tie(w,result) = ColorfulGraphMotifInstance(G, M, k, LL).solve();

    if(result.size() > w_size || (result.size() >= w_size && w>w_opt)) {
      w_opt = w;
      w_size = result.size();
      r_opt = result;
    }
  }
  cout << "solution weight           : " << w_opt << endl;
  return r_opt;
}
