/*   Gram -- A graph motif solver.
 *
 *   (C) 2008-2009 René van Bevern <m5bere2@uni-jena.de>
 *
 *   This file is part of Gram.
 *
 *   Gram is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Gram is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Gram.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/function.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_int.hpp>
#include "GraphUtils.hpp"
#include "LUT.hpp"

typedef unordered_map<Color, ColorSet> ColorSets;

/** find colorful graph motives in graphs */
class ColorfulGraphMotifInstance {
protected:
  typedef vector<ColorSet> Reps;
  
  /** colorful motif being sought */
  ColorSet M;
  ColorSet L;

  /** size of colorful motif being sought */
  int k;

  /** current mapping of vertices to vertex colors */
  const VertexColors& coloring;
  const VertexColorLists& lists;

  /** lookup table for dynamic programming. for each vertex @a v and
      each colorset @a C stores, if v is part of an occurence of @a C
      in #G. */
  LUT table;

  /** graph that is searched for the motif */
  Graph G;

  /** check if a vertex can be added to existing solutions.

      @pre #table has to store the correct
      information for all neighbors of @a v and all subsets of @a C.

      @post if @a v is part of @a C is stored in #table.

      @return it is returned if @a v is part of @a C. */
      
  int growMotifs(Vertex v, const ColorSet& C);

  /** check if two two existing solutions can be joined.
      @pre #table has to store the correct
      information for all neighbors of @a v and all subsets of @a C.
      
      @post if @a v is part of @a C is stored in #table.
      
      @return it is returned if @a v is part of @a C. */
  bool joinMotifs(Vertex v, const ColorSet& C, int thres);
public:

  /** find graph motif @return a solution of size #k if an instance of
      #M can be found in #G, otherwise an empty solution.  @post if
      DESTRUCTIVE is defined at time of compilation,
      all vertices that are not relevant for a solution are deleted in
      each iteration step.  A vertex is relevant for a solution of
      size @a k if it part of a solution of size @a (k-1) */
  tuple<VertexWeight, Solution> solve();

  /** Construct colorful graph motif instance. The following
   * preprocessing steps are done here:
      - initialization of the backtracking table
      - removal of isolated vertices if @a k is greater than 1
      - removal of unicolored edges

      @param G graph to search motif in
      @param M motif an instance is to be found of in @a G
      @param k size of motif instance to be found
  */
      
  ColorfulGraphMotifInstance(Graph& G, ColorSet M, int k, ColorSet L);
};

/** find general graph motives in graphs */
class GraphMotifInstance {
protected:
  /** colorful motif being sought */
  ColorSet M;

  /** size of colorful motif being sought */
  int k;

  /** current mapping of vertices to vertex colors */
  VertexColors coloring;

  /** graph that is searched for the motif */
  Graph G;

  ColorPoolReps L;
  vector<set<Vertex> > Ls;
  vector<int> colorsPerL;
  
  float errorProb;
  typedef mt19937 generator_type;
  typedef uniform_int<int> distribution;
  typedef variate_generator<generator_type&, distribution> generator;
  
  generator_type RNG;

  int getRand(int max) {
    return generator(RNG, distribution(0, max - 1))();
  }

  /** recolor the graph.  @post the vertices in #toRecolor are
      destructively given new colors from #pool. */
  void recolor();
public:
  /** reseed the internal random generator */
  void reseed() { RNG.seed(RNG()); }
  
  /** construct graph motif instance. The following preprocessing is done in this step:
      - vertices with wrong color are deleted
      - isolated vertices are removed if M has more than one element
      @param G graph to search motif in
      @param M motif an instance is to be found of in @a G
      @param err the probability with which the algorithm may not find a
      solutuion. This value defaults to 0.1.
  */
  GraphMotifInstance(const Graph& G, const MultiColorSet& M, float err = 0.1);

  /** find graph motif @return a solution of size #k if an instance of
      #M can be found in #G, otherwise #recolor #G and try again, until
      a certain number of ::trials is exceeded. */
  Solution solve();
};
